/**
 * \file SimpleRemote.c
 *
 * \brief SimpleRemote application
 *
 * Copyright (C) 2012-2013, Atmel Corporation. All rights reserved.
 *
 * \asf_license_start
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * 3. The name of Atmel may not be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * 4. This software may only be redistributed and used in connection with an
 *    Atmel microcontroller product.
 *
 * THIS SOFTWARE IS PROVIDED BY ATMEL "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT ARE
 * EXPRESSLY AND SPECIFICALLY DISCLAIMED. IN NO EVENT SHALL ATMEL BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * \asf_license_stop
 *
 * Modification and other use of this code is subject to Atmel's Limited
 * License Agreement (license.txt).
 *
 * $Id: SimpleRemote.c 9267 2014-03-18 21:46:19Z ataradov $
 *
 */

/*- Includes ---------------------------------------------------------------*/
#include <string.h>
#include "config.h"
#include "hal.h"
#include "phy.h"
#include "sys.h"
#include "nwk.h"
#include "halGpio.h"
#include "halUart.h"
#include "dsp.h"
#include "data.h"

/*- Definitions ------------------------------------------------------------*/
HAL_GPIO_PIN(LED, A, 14);
HAL_GPIO_PIN(BUTTON, A, 15);

#define APP_BUFFER_SIZE     NWK_MAX_PAYLOAD_SIZE

/*- Types ------------------------------------------------------------------*/
typedef enum AppState_t
{
	APP_STATE_INITIAL,
	APP_STATE_PROCESS,
	APP_STATE_TRANSMIT,
	APP_STATE_WAIT_CONF,
	APP_STATE_PREPARE_DATA,
	APP_STATE_PREPARE_SLEEP,
	APP_STATE_SLEEP,
	APP_STATE_IDLE,
} AppState_t;

/*- Variables --------------------------------------------------------------*/
static AppState_t			appState = APP_STATE_INITIAL;
static bool					appBtnState = true;

static NWK_DataReq_t		appDataReq;
static bool					appDataReqBusy = false;
static uint8_t				appDataReqBuffer[APP_BUFFER_SIZE];

static uint8_t				appDataTxBuffer[APP_BUFFER_SIZE];
static uint8_t				appDataTxSize;
static uint8_t				appDataLoops;

static uint16_t				appSampleIterations;
static uint16_t				appSamplesProcessed;
static uint16_t				appSamplesTodo;
static uint16_t				appSamplesPerPacket;
static uint16_t				appSamplesRemaining;

//	Variables for ADPCM encoder
#ifdef APP_PROCESS_ADPCM
	static int16_t			stepIndex;
	static int16_t			predictedValue;
#endif

/*- Prototypes -------------------------------------------------------------*/
static bool appSendData(uint8_t *data, uint8_t length);
static void APP_TaskHandler(void);

/*- Implementations --------------------------------------------------------*/
int main(void)
{
	SYS_Init();
	
	while (1)
	{
		SYS_TaskHandler();
		APP_TaskHandler();
	}
}

//	Confirm data has been transmitted
static void appDataConf(NWK_DataReq_t *req)
{
	appDataReqBusy = false;
	appState = APP_STATE_PREPARE_DATA;
	(void)req;
}

//	Send data
static bool appSendData(uint8_t *data, uint8_t length)
{
	if (appDataReqBusy)
	return false;

	memcpy(appDataReqBuffer, data, length);

	appDataReq.dstAddr = 1-APP_ADDR;
	appDataReq.dstEndpoint = APP_ENDPOINT;
	appDataReq.srcEndpoint = APP_ENDPOINT;
	appDataReq.options = NWK_OPT_ACK_REQUEST;
	appDataReq.data = appDataReqBuffer;
	appDataReq.size = length;
	appDataReq.confirm = appDataConf;
	NWK_DataReq(&appDataReq);

	appDataReqBusy = true;
	return true;
}

//	Handle incoming data
static bool appDataInd(NWK_DataInd_t *ind)
{
	return true;
	(void)ind;
}

//	Check button state
static bool checkButton(void)
{
	bool b = HAL_GPIO_BUTTON_read();
	
	if (b != appBtnState)
	{
		appBtnState = b;
		return !b;
	}
	else return false;
}

//	Initialize application
static void appInit(void)
{
	HAL_GPIO_BUTTON_in();
	HAL_GPIO_BUTTON_pullup();
	
	HAL_GPIO_LED_out();
	HAL_GPIO_LED_set();
	
	NWK_SetAddr(APP_ADDR);
	NWK_SetPanId(APP_PANID);
	PHY_SetChannel(APP_CHANNEL);
	//PHY_SetTxPower(0x0F);				//	-17 dBm
	PHY_SetTxPower(0x00);				//	+3 dBm
	PHY_SetRxState(true);
	
	NWK_OpenEndpoint(APP_ENDPOINT, appDataInd);
	
	if (!NWK_Busy())
	{
		NWK_SleepReq();
	}
}

//	Initialize data
static void appDataInit(void)
{
	#if defined(APP_PROCESS_CP)
		appSamplesPerPacket = (APP_BUFFER_SIZE / 2);
	#elif defined(APP_PROCESS_ADPCM)
		appSamplesPerPacket = (APP_BUFFER_SIZE * 2);
		stepIndex = 0;
		predictedValue = 0;
	#endif
	
	appDataLoops = APP_DATA_LOOPS;
}

static void appDataLoopInit(void)
{
	appSampleIterations = APP_NUM_SAMPLES / appSamplesPerPacket;
	appSamplesRemaining = APP_NUM_SAMPLES % appSamplesPerPacket;
	appSamplesProcessed = 0;
	if (APP_NUM_SAMPLES >= appSamplesPerPacket) appSamplesTodo = appSamplesPerPacket;
	else appSamplesTodo = appSamplesRemaining;
}

//	Application task handler
static void APP_TaskHandler(void)
{
	switch (appState)
	{
		case APP_STATE_INITIAL:
			appInit();
			appDataInit();
			appDataLoopInit();
			appState = APP_STATE_IDLE;
		break;

		case APP_STATE_IDLE:
			if (checkButton())
			{
				NWK_WakeupReq();
				appState = APP_STATE_PROCESS;
			}
		break;
		
		case APP_STATE_PROCESS:
			#if defined(APP_PROCESS_CP)
				appDataTxSize = appSamplesTodo * 2;
				memcpy(appDataTxBuffer, &srcSamples[appSamplesProcessed], appDataTxSize);
			#endif
			#if defined(APP_PROCESS_ADPCM)
				appDataTxSize = appSamplesTodo / 2;
				dsp_adpcm_ima_encode(appDataTxBuffer, &srcSamples[appSamplesProcessed], appSamplesTodo, &stepIndex, &predictedValue);
			#endif
			appSamplesProcessed += appSamplesTodo;
			if (appSampleIterations > 0) appSampleIterations--;
			appState = APP_STATE_TRANSMIT;
			//appState = APP_STATE_PREPARE_DATA;
		break;
		
		case APP_STATE_TRANSMIT:
			if (appSendData(appDataTxBuffer, appDataTxSize))
			{
				appState = APP_STATE_WAIT_CONF;
			}
		break;
		
		case APP_STATE_PREPARE_DATA:
			if (appSampleIterations == 0)
			{
				if (appSamplesProcessed < APP_NUM_SAMPLES)
					appSamplesTodo = appSamplesRemaining;
				else
				{
					appDataLoops--;
					if (appDataLoops == 0)
					{
						appState = APP_STATE_PREPARE_SLEEP;
					}
					else
					{
						appDataLoopInit();
						appState = APP_STATE_PROCESS;
					}
					break;
				}
			}
			appState = APP_STATE_PROCESS;
		break;
		
		case APP_STATE_PREPARE_SLEEP:
			if (!NWK_Busy())
			{
				NWK_SleepReq();
				appState = APP_STATE_SLEEP;
				SCB->SCR |= SCB_SCR_SLEEPDEEP_Msk;
				__DSB();
				__WFI();
				//HAL_GPIO_LED_clr();
			}
		break;

		default:
		break;
	}
}